/* Copyright (C) 2015, 2016 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package sjc.debug;

import sjc.compbase.AccVar;
import sjc.compbase.Context;
import sjc.compbase.Modifier;
import sjc.compbase.Mthd;
import sjc.compbase.Unit;
import sjc.compbase.Vrbl;
import sjc.memory.ImageContainer;
import sjc.memory.MemoryImage;
import sjc.memory.MemoryObjectDebugInfo;
import sjc.osio.TextPrinter;

/**
 * AsmOutX86: symbol information writer for x86 assembler output
 * 
 * @author S. Frenz
 * @version 160201 fixed scalar size for objects of unknown type, now explicitly extending unicode strings
 *  version 160109 assemble-ready reloc/scalar/string distinctive output
 *  version 160102 added support for structured unit output
 *  version 151108 initial version
 */

public class AsmOutX86 extends DebugLister {
  private Context ctx;
  private TextPrinter out;
  private ImageContainer img;
  
  public AsmOutX86(String filename, Context ictx) {
    out=(ctx=ictx).osio.getNewFilePrinter(filename);
  }
  
  public void globalRAMInfo(boolean isDecompressor, Object ramInitLoc, int ramSize, int constMemorySize) {
    if (isDecompressor) out.println(";--- Decompressor ---");
  }
  
  public void listMemory(MemoryImage mem) {
    Unit unit;
    Mthd mthd;
    String string;
    
    if (ctx.mem instanceof ImageContainer) img=(ImageContainer)ctx.mem;
    else {
      out.println("memory error, see compilation output");
      ctx.out.println("AsmOutX86 debug lister: memory is not readable, can not dump it");
      return;
    }
    
    out.println(";--- assembler output generated by SJC ---");
    out.print("BITS ");
    out.println(ctx.arch.relocBytes*8);
    out.print("ORG 0x");
    out.printHexFix(mem.getBaseAddress(), 8);
    out.println();
    out.println();
    
    MemoryObjectDebugInfo now=mem.getFirstDebugListerObject();
    while (now!=null) {
      if (now.source instanceof Unit) {
        unit=(Unit)now.source;
        out.print(";--- start of class desc for unit ");
        unit.printNameWithOuter(out);
        out.println();
        printSClassDescRelocs(now.pointer, now.relocs, unit);
        unit.printNameWithOuter(out);
        out.println("$clss:");
        printSClassDescScalars(now.pointer, now.scalarSize, unit);
      }
      else if (now.source instanceof Mthd) {
        mthd=(Mthd)now.source;
        out.print(";--- start of method ");
        out.print(mthd.owner.name);
        out.print(".");
        mthd.printNamePar(out);
        out.println();
        printRelocs(now.pointer, now.relocs);
        printObjectScalars(now.pointer, ctx.codeStart, ctx.langRoot, false);
        out.print(mthd.asmCode);
        out.print("    TIMES ");
        out.print(now.scalarSize);
        out.print("-$+");
        printRelocLabel(now.pointer);
        out.println(" db 0");
      }
      else if (now.source instanceof String) {
        string=(String)now.source;
        out.println(";--- start of String inst");
        printRelocs(now.pointer, now.relocs);
        if (ctx.langString.inlArr!=null) {
          printObjectScalars(now.pointer, ctx.langString.instScalarTableSize, ctx.langString, false);
          printEscapedString(string, now.scalarSize-ctx.langString.instScalarTableSize);
        }
        else {
          printObjectScalars(now.pointer, now.scalarSize, ctx.langString, false);
          now=now.next; //next object is char-array that belongs to given string
          printRelocs(now.pointer, now.relocs);
          printObjectScalars(now.pointer, ctx.rteSArray.instScalarTableSize, ctx.rteSArray, false);
          printEscapedString(string, now.scalarSize-ctx.rteSArray.instScalarTableSize);
        }
      }
      else {
        //if (now.source!=null) System.out.println("missing special implementation for now.source type "+now.source.getClass().getName());
        out.println(";--- start of common object");
        printRelocs(now.pointer, now.relocs);
        printObjectScalars(now.pointer, now.scalarSize, null, false);
      }
      now=now.next;
    }
  }
  
  public void finalizeImageInfo() {
  }
  
  private int getMemValue(int addr, int size) {
    int value=0;
    int shift=0;
    while (size-->0) {
      value|=((int)img.memBlock[addr-img.baseAddress]&0xFF)<<shift;
      addr++;
      shift+=8;
    }
    return value;
  }
  
  private void printGeneralObjPtr(int pointer) {
    printRelocLabel(pointer);
    out.println(":");
  }
  
  private void printRelocLabel(int pointer) {
    if (pointer!=0) {
      out.print("$_A_");
      out.printHexFix(pointer, 8);
    }
    else out.print("0");
  }
  
  private void printRelocs(int pointer, int relocs) {
    int i;
    for (i=relocs; i>0; i--) printReloc(pointer-i*ctx.arch.relocBytes);
  }
  
  private void printReloc(int pointer) {
    switch (ctx.arch.relocBytes) {
      case 1:
        out.print("  db ");
        break;
      case 2:
        out.print("  dw ");
        break;
      case 4:
        out.print("  dd ");
        break;
      case 8:
        out.print("  dq ");
        break;
    }
    printRelocLabel(getMemValue(pointer, ctx.arch.relocBytes));
    out.print("   ;");
    out.printHexFix(pointer, 8);
    out.println();
  }
  
  private void printSClassDescRelocs(int pointer, int relocs, Unit unit) {
    int i, off;
    AccVar var;
    Mthd mthd;
    for (i=relocs; i>0; i--) {
      off=-i*ctx.arch.relocBytes;
      var=searchVar(unit, AccVar.L_CLSSREL, off, false);
      if (var==null) var=searchVar(ctx.rteSClassDesc, AccVar.L_INSTREL, off, true);
      if (var!=null) {
        ctx.printUniqueVarName(out, var, unit);
        out.println(":");
      }
      else {
        mthd=searchMthd(unit, off);
        if (mthd!=null) {
          ctx.printUniqueMethodName(out, mthd, unit);
          out.println("$ref:");
        }
      }
      printScalar(pointer+off, ctx.arch.relocBytes);
    }
  }
  
  private void printObjectScalars(int pointer, int scalarSize, Unit unit, boolean withNames) {
    AccVar var=null;
    int off=0, size;
    while (off<scalarSize) {
      size=1;
      if (unit!=null) var=searchVar(unit, AccVar.L_INSTSCL, off, true);
      if (var!=null) {
        if (withNames) {
          ctx.printUniqueVarName(out, var, unit);
          out.println(":");
        }
        size=var.minSize;
      }
      printScalar(pointer+off, size);
      off+=size;
    }
  }
  
  private void printSClassDescScalars(int pointer, int scalarSize, Unit unit) {
    AccVar var;
    int off=0, size;
    while (off<scalarSize) {
      size=1;
      var=searchVar(unit, AccVar.L_CLSSSCL, off, false);
      if (var==null) var=searchVar(ctx.rteSClassDesc, AccVar.L_INSTSCL, off, true);
      if (var!=null) {
        ctx.printUniqueVarName(out, var, unit);
        out.println(":");
        size=var.minSize;
      }
      printScalar(pointer+off, size);
      off+=size;
    }
  }
  
  /*private void printInlineArrayScalars(int pointer, int scalarSize, Unit unit) {
    int entrySize=unit.inlArr.minSize;
    int entries=(scalarSize-unit.instScalarTableSize)/entrySize;
    for (int i=0; i<entries; i++) printValue(pointer, unit.instScalarTableSize+i*entrySize, entrySize);
  }*/
  
  private void printEscapedString(String what, int desiredByteLength) {
    int i;
    if (ctx.byteString) out.print("  db ");
    else out.print("  dw ");
    for (i=0; i<what.length(); i++) {
      if (i>0) out.print(',');
      char c=what.charAt(i);
      if (c<32 || c>127 || c=='\'' || c=='\\') {
        out.print("0x");
        out.printHexFix(c, 2);
      }
      else {
        out.print('\'');
        out.print(c);
        out.print('\'');
      }
      if (ctx.byteString) desiredByteLength--;
      else desiredByteLength-=2;
    }
    out.println();
    if (desiredByteLength>0) {
      out.print("  db ");
      i=0;
      while (i<desiredByteLength) {
        if (i>0) out.print(',');
        out.print('0');
        i++;
      }
      out.println();
    }
  }
  
  private void printScalar(int pointer, int size) {
    if (size>ctx.arch.relocBytes) { //too big for single access, split into multiple smaller portions
      while (size>0) {
        printScalar(pointer, ctx.arch.relocBytes);
        pointer+=ctx.arch.relocBytes;
        size-=ctx.arch.relocBytes;
      }
      return;
    }
    printGeneralObjPtr(pointer);
    switch (size) {
      case 1:
        out.print("  db 0x");
        break;
      case 2:
        out.print("  dw 0x");
        break;
      case 4:
        out.print("  dd 0x");
        break;
      case 8:
        out.print("  dq 0x");
        break;
    }
    out.printHexFix(getMemValue(pointer, size), size*2);
    out.println();
  }
  
  private Mthd searchMthd(Unit unit, int offset) {
    Mthd mthds;
    boolean stat=true;
    while (unit!=null) {
      mthds=unit.mthds;
      while (mthds!=null) {
        if (mthds.relOff==offset && (stat || (mthds.modifier&Modifier.M_STAT)==0)) return mthds;
        mthds=mthds.nextMthd;
      }
      if (unit.extsID!=null) unit=unit.extsID.unitDest;
      else unit=null;
      stat=false;
    }
    return null;
  }
  
  private AccVar searchVar(Unit unit, int reqLocation, int offset, boolean searchParents) {
    Vrbl vars;
    while (unit!=null) {
      vars=unit.vars;
      while (vars!=null) {
        if (vars.location==reqLocation && vars.relOff==offset) return vars;
        vars=vars.nextVrbl;
      }
      if (searchParents && unit.extsID!=null) unit=unit.extsID.unitDest;
      else unit=null;
    }
    return null;
  }
}
